@@ -70,7 +70,7 @@ def extract(path,variables=None,verbose=False,all_variables=None,playwright=Fals
         m = comment_pat.match(line)
         if not m:
             m = I18NOK_pat.search(line)
-        if m or not variable_section or line=="\n" or is_selector_class(line):
+        if m or not variable_section or line=="\n" or line == "\r\n" or is_selector_class(line):
             robotfile += line  # just copy line
             continue
 
@@ -82,7 +82,7 @@ def extract(path,variables=None,verbose=False,all_variables=None,playwright=Fals
                 if var["key"] not in keys:
                     propfile += f"{var['key']} = {var['value']}\n"
                     keys[var['key']] = 0
-                    keys[var['key']] += 1
+                keys[var['key']] += 1
             return propfile
 
         if playwright and is_selector_pw(line):
@@ -270,7 +270,11 @@ def is_non_localizable(value):
 text_pat = re.compile(r'''(?P<oper>(contains|starts-with)\s*\(\s*)?(?P<cclass>text\(\)\s*|@aria-label\s*|@placeholder\s*|@title\s*|\.\s*?)(?P<dlm>[=,]\s*)(['"])(?P<T>.*?)\5(?P<oper_end>\s*\))?''')
 pw_pat = re.compile(r'''(?P<oper>\[\s*\(?\s*)?(?P<cclass>text\s*|aria-label\s*|placeholder\s*|@title\s*|\.\s*?)(?P<dlm>[=]\s*)(['"])(?P<T>.*?)\4(?P<oper_end>\s*\)?\s*\])?''')
 pwprop_pat = re.compile(r'''\[(id|aria-label|placeholder|text)\s*=.*\]|text\s*=|css\s*=|:has-text\s*\(''')
-def extract_selector_xpath(line,variables,all_variables,verbose=None):
+def extract_selector_xpath(line,variables, all_variables, verbose=None, locale=None, add_english=False,fuzzy_match=False):
+    langlist = None
+    if locale:
+        langlist = []
+        langlist.append(locale)
     m = pwprop_pat.search(line)  # is playwright selector
     if m:
         seg = pw_pat.finditer(line)   # playwright xpath
@@ -280,10 +284,11 @@ def extract_selector_xpath(line,variables,all_variables,verbose=None):
     if seg:
         count = 0
         lastchar = 0
-        props = []
+        props = [] # initialize per line
         vars = []
         edlist = []
         for m in iter(seg):
+            prop = []  # initialize props per value
             strcls = m.group('cclass')
             #print (f"strcls={strcls}")
             value = m.group('T')
@@ -293,29 +298,46 @@ def extract_selector_xpath(line,variables,all_variables,verbose=None):
 
             #newline = newline[:m.start()] + m.group(1) + m.group(2) + "'" + var + "'" + newline[m.end():]
             if variables:
-                props.extend(bundle_reference(value,variables))
+                prop = bundle_reference(value,variables)
+                props.extend(prop)
 
             value0 , nhit = var_reference(value,variables,hint=strcls,verbose=verbose)
             if nhit:
-                # print (f"Indirect reference: {var} = {value}")
+                #print (f"Indirect reference: {var} = {value}")
                 var = value0
+                # if fuzzy_match:
+                #     print (f"var_reference: var={var} value0={value0} nhit={nhit}")
+            elif fuzzy_match:
+                value0, nhit = var_fuzzy_match(value,variables,verbose=verbose)
+                if nhit:
+                    # print (f"var_fuzzy_match: var={var} value0={value0} nhit={nhit}")
+                    nhit = 0  # reset nhit not to call detect_varint_trans
+                    #var = value0
+                    var = value0 + "*"  # append "* to indicate fuzzy matching
+                else:
+                    vars.append({ "key":var, "value":value })  # unresolved in fuzzy match
             else:
                 vars.append({ "key":var, "value":value })
 
             varlist = None
             if all_variables and nhit:
-                varlist = detect_variant_trans(var,props,all_variables,hint=strcls)
+                #print(f"Debug0: nhit={nhit} {varlist} {var} props={props}")
+                varlist = detect_variant_trans(var,prop,all_variables,hint=strcls,langlist=langlist)
 
             edlist.append({"begin":m.start(), "end":m.end(),"var":var if not varlist else varlist,
                 "g1":m.group('cclass'), "g2":m.group('dlm'),
-                "oper":m.group('oper'), "oper_end":m.group('oper_end')})
+                "oper":m.group('oper'), "oper_end":m.group('oper_end'), "value":value})
             count += 1
 
         if count == 0:
             return True, line, props, vars   #return True and discad untranslatable patterns
 
         edlist.reverse()
-        line = edit_line(line,edlist)
+        if locale:
+            line = edit_line(line,edlist,all_variables,new_variables=vars,locale=locale,add_english=add_english)
+            # resolve variables for locale
+        else:
+            line = edit_line(line,edlist)
 
     else:
         print (f"Return from finditer is None: {line}")
@@ -348,14 +370,13 @@ def varname(label):
 def var_reference(label,variables=None,hint=None,verbose=False):
     # hint: text(), ., @placeholder, @aria-label, @title
     # title does not mean <title>, but title= attribute.
-    chrclass = { "placeholder":[], "label":[], "title":[], "tooltip":[], "icon":[] }
+    chrclass = { "placeholder":[], "label":[], "title":[], "tooltip":[], "icon":[], "prefix":[] }
     nhit = 0
     candidates = []
 
     #
     debug = False
-    if label == "Model Attestations":
-        debug = True
+
     if not variables:
         return label, 0
 
@@ -371,6 +392,11 @@ def var_reference(label,variables=None,hint=None,verbose=False):
     if len(candidates)==1:  # only one found
         label = candidates[0]
         return label, 1
+
+    # if multiple hit
+    if verbose:
+        print(f"candidates={candidates}")
+
     # clasify keys
     for key in candidates:
         lkey = key.lower()
@@ -384,6 +410,8 @@ def var_reference(label,variables=None,hint=None,verbose=False):
             chrclass["label"].append(key)
         elif lkey.find("tooltip")>=0:
             chrclass["tooltip"].append(key)
+        elif lkey.find("prefix")>=0:
+            chrclass["prefix"].append(key)
         else:
             chrclass["label"].append(key)
 
@@ -411,6 +439,10 @@ def var_reference(label,variables=None,hint=None,verbose=False):
         label0, count0 = select("label",candidates)
         if count0 > 0:
             return label0, count0
+    elif hint == "prefix":
+        label0, count0 = select("prefix",candidates)
+        if count0 > 0:
+            return label0, count0
 
     label0, count0 = select("label",candidates)
     if count0 > 0:
@@ -463,8 +495,9 @@ def write_file(path,buffer,destname):
     print (f"{nlines} lines written to \"{path}\".")
     return 1
 
-def detect_variant_trans(def_key,props,all_variables,hint=None):
+def detect_variant_trans(def_key,props,all_variables,hint=None,langlist=None):
     # return list of variables if variant translation found
+    #print(f"Debug: def_key={def_key} props={props} ")
     if not all_variables:
         return None
     count = 0
@@ -472,23 +505,59 @@ def detect_variant_trans(def_key,props,all_variables,hint=None):
     uniq_trans = {}
 
     uniq_keys.append(def_key)
-    uniq_trans[all_variables["variables"]['en'][def_key]] = 1
+    english_string = all_variables["variables"]['en'][def_key]
+    uniq_trans[english_string] = 1   # register EN string
+    props.insert(0,def_key)
 
     for key in props:
         for loc in all_variables["locales"]:
+            if langlist and loc not in langlist:
+                continue  # skipt language
             if all_variables["variables"][loc][key] not in uniq_trans:
                 uniq_trans[all_variables["variables"][loc][key]] = 0
                 if key not in uniq_keys:
                     uniq_keys.append(key)
-            uniq_trans[all_variables["variables"][loc][key]] += 1
+
+                uniq_trans[all_variables["variables"][loc][key]] += 1
     count = len(uniq_keys)
     if count>1:
-        print(f"varinats found: ={uniq_keys}")
+        print(f"variants found: ={uniq_keys}")
+        print(f"\t{uniq_trans}")
         return uniq_keys
     return None
 
-def edit_line(line,edlist):
+def edit_line(line,edlist,all_variables=None,new_variables=None,locale=None,add_english=False):
+    # print (f"Debug(edlilne):\t{line}\n\t{edlist}")
     newline = line
+    variables = []
+    loc_variables = []
+    new = {}
+    if all_variables and locale:
+        variables = all_variables["variables"]["en"]
+        loc_variables = all_variables["variables"][locale]
+    if new_variables:
+        for va in new_variables:
+            new[va["key"]] = va["value"]
+
+    def resolve_var(var,value,variables,loc_variables):
+        if variables:
+            if var.endswith("*"): # fuzzy matching
+                resolved = resolve_fuzzy_variable(var,value,variables,loc_variables)
+                var = '"' + resolved + '"'
+                return var
+
+            if var in loc_variables:
+                var = '"' + loc_variables[var] + '"'
+            elif var in variables:
+                var = '"' + variables[var] + '"'
+            elif new and var in new:
+                var = '"' + new[var] + '"'
+            else:
+                print (f"Warning: Unresolved variable: {var}")
+                var = '"${' + var + '}"'
+        else:
+            var = '"${' + var + '}"'
+        return var
 
     for edstr in edlist:
     #newline = newline[:ix["begin"]] + ix["g1"] + ix["g2"] + '"${' + ix["var"] + '}"' + newline[ix["end"]:]
@@ -505,26 +574,45 @@ def edit_line(line,edlist):
             outer_op = edstr["oper"]
         elif edstr["oper_end"]:
             outer_cl = edstr["oper_end"]
+        value = edstr["value"]
 
-        if isinstance(edstr["var"],list):  # multiple translations
+        # pre-process edstr["var"]
+        vars = edstr["var"]
+        if isinstance(vars,list):
+            vars = []
+            for var in edstr["var"]:
+                vars.append(resolve_var(var,value,variables,loc_variables))
+            if add_english:
+                if edstr["var"][0] in variables:
+                    vars.append(f'"{variables[edstr["var"][0]]}"')  # append English string
+        elif add_english:
+            var = vars
+            vars = []
+            vars.append(resolve_var(var,value,variables,loc_variables))
+            if var in variables:
+                vars.append(f'"{variables[var]}"')
+        else:
+            vars = resolve_var(vars,value,variables,loc_variables)
+
+        if isinstance(vars,list):  # multiple translations
             # ( @title = "${var[0]}" or $title = "${var[1]} ... ")
             # contains(@title,"${var[0]}" or )
             condline = "("
             orstr = ""
-            for var in edstr["var"]:
-                ref = openstr + edstr["g1"] + edstr["g2"] + '"${' + var + '}"' + closestr
+            #edit(f"Debug: vars={vars}")
+            for var in vars:
+                ref = openstr + edstr["g1"] + edstr["g2"] + var + closestr
                 condline += orstr + ref
                 orstr = " or "
             condline += ")"
             ref = outer_op + condline + outer_cl
-            print (f"ref={ref}")
+            # print (f"ref={ref}")
         else:
-            ref = outer_op + openstr + edstr["g1"] + edstr["g2"] + '"${' + edstr["var"] + '}"' + closestr + outer_cl # @title = "${var}"
+            ref = outer_op + openstr + edstr["g1"] + edstr["g2"] + vars + closestr + outer_cl # @title = "${var}"
         newline = newline[:edstr["begin"]] + ref + newline[edstr["end"]:]
 
     # print(f"edit_line:outer_op={outer_op} openstr={openstr} closestr={closestr} outer_cl={outer_cl}")
     # sys.stdout.write('"%s"\n"%s"\n' % (line.strip('\n'),newline.strip('\n')))
-    newline = newline.rstrip()  # remove trailing spaces
 
     return newline
 
@@ -537,6 +625,104 @@ def is_placeholder(value):
         return True
     return False
 
+# exracor for XPATH file
+def extract_xpath(path,variables=None,verbose=False,all_variables=None,playwright=False,fuzzy_match=False):
+    DIAG = False
+    propfile = ""
+    syspropfile = ""
+    sysprop = []
+
+    if not os.path.exists(path):
+        print (f"{path} does not exist.")
+        return None, None
+
+    with codecs.open(path,'r',encoding='utf-8') as fp:
+        lines = fp.readlines()
+
+    keys = {}  # store new keys
+    variable_section = False
+    nxpath = 0
+    for line in lines:
+        line = line.strip('\n')
+        # check whether a xpath contains localizalbe string
+
+        m = text_pat.search(line)
+        tx_sw = bool(m)
+        if DIAG:
+            print (f"{tx_sw}: {line}")
+        if not tx_sw:
+            continue
+        found, line2, props , vars = extract_selector_xpath(line,variables,all_variables,verbose=verbose,fuzzy_match=fuzzy_match)
+        if not found:
+            print (f"!{line2}")
+        nxpath += 1
+        # if nxpath == 24:
+        #     print (f"{found} line2={line2}\n\tprops={props} vars={vars}")
+        if found:
+            #syspropfile += update_props_vars(line2,nxpath,props,vars)
+            syspropfile += f"xpath{nxpath} = {line}\n"
+            for var in vars:
+                if var["key"] not in keys:
+                    keys[var["key"]] = 0
+                    propfile += f'{var["key"]} = {var["value"]}\n'
+                keys[var["key"]] += 1
+        else:
+            print (f"No translatable string: {line2}")
+
+    return syspropfile, propfile
+
+namevalue = re.compile(r'(?P<N>xpath\d+) = (?P<V>.*)')
+def localize_xpath(lines0, all_variables, locale=None, add_english=False,fuzzy_match=False):
+    propfile = ""
+    variables = all_variables["variables"]["en"]
+    lines = lines0.splitlines()
+    for line in lines:
+        m = namevalue.match(line)
+        if not m:
+            propfile += f"{line}\n"
+        else:
+            name = m.group('N')
+            value = m.group('V')
+            found, translated, props , vars = extract_selector_xpath(value,variables,all_variables,
+                verbose=False,locale=locale,add_english=add_english,fuzzy_match=fuzzy_match)
+            #print(f"{found} props={props} vars={vars}\n translated={translated}")
+            propfile += f"{name} = {translated}\n"
+    return propfile
+
+def var_fuzzy_match(label,variables=None,verbose=False):
+    #
+    # neee to handle multiple matches
+    #
+    value = label
+    nhit = 0
+    tokens = label.split(' ')
+    if len(tokens)<2:  # only one token
+        return value, 0
+    var = ' '.join(tokens[0:-1])
+    print (f"tokens={tokens} token={var}")
+    value0 , nhit = var_reference(var,variables,hint="prefix",verbose=True)
+    #print (f"\tvalue0={value0} value={value} nhit={nhit}")
+    if nhit>0:
+        return value0, nhit
+    return value, nhit
+
+def resolve_fuzzy_variable(key,value,variables,loc_variables):
+
+    tokens = value.split(' ')
+    if key.endswith("*"):
+        key = key[0:-1]  # strip trailing '*'
+    if key.endswith(".prefix"):
+        key = key[0:-(len(".prerix"))]  # need to resolve placeholder
+        value = loc_variables[key]
+    else:
+        value = loc_variables[key] + ' {0}'
+
+    #tokens = variables[key].split(' ')
+    print(f"resolve_fuzzy_variable: key={key} value={value} tokens={tokens}")
+    value = re.sub(r'{\d}',tokens[-1],value)
+
+    return value
+
 if __name__ == "__main__":
 
     text ='''${Obj_DE_ZeroState_Svg}    xpath://*[text()= "No data is selected."] '''
